/*******************************************************************************
 * Copyright (c) 2009, 2010 Mountainminds GmbH & Co. KG and Contributors
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Marc R. Hoffmann - initial API and implementation
 *    
 * $Id: $
 *******************************************************************************/
package org.jacoco.core.data;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * Additional data output methods for compact storage of data structures.
 * 
 * @see CompactDataOutput
 * @author Marc R. Hoffmann
 * @version $Revision: $
 */
public class CompactDataOutput extends DataOutputStream {

	/**
	 * Creates a new {@link CompactDataOutput} instance that writes data to the
	 * specified underlying output stream
	 * 
	 * @param out
	 *            underlying output stream
	 */
	public CompactDataOutput(final OutputStream out) {
		super(out);
	}

	/**
	 * Writes a variable length representation of an integer value that reduces
	 * the number of written bytes for small positive values. Depends on the
	 * given value 1 to 5 bytes will be written to the underlying stream.
	 * 
	 * @param value
	 *            value to write
	 * @throws IOException
	 */
	public void writeVarInt(final int value) throws IOException {
		if ((value & 0xFFFFFF80) == 0) {
			writeByte(value);
		} else {
			writeByte(0x80 | (value & 0x7F));
			writeVarInt(value >>> 7);
		}
	}

	private int booleanBuffer = 0;

	private int booleanBufferSize = 0;

	/**
	 * Writes a boolean value. Internally a sequence of boolean values is packed
	 * into single bits. After the last boolean value has been written
	 * {@link #finishPackedBoolean()} has to be called.
	 * 
	 * @param value
	 *            boolean value
	 * @throws IOException
	 */
	public void writePackedBoolean(final boolean value) throws IOException {
		if (value) {
			booleanBuffer |= 0x01 << booleanBufferSize;
		}
		if (++booleanBufferSize == 8) {
			finishPackedBoolean();
		}
	}

	/**
	 * Finalizes the output of a sequence of packed boolean values.
	 * 
	 * @throws IOException
	 */
	public void finishPackedBoolean() throws IOException {
		if (booleanBufferSize > 0) {
			writeByte(booleanBuffer);
			booleanBuffer = 0;
			booleanBufferSize = 0;
		}
	}

}
